from typing import Optional, Any, List, TypeVar, Type, cast, Callable

from topgreenerapi.capability import Capability

T = TypeVar("T")


def from_str(x: Any) -> str:
    assert isinstance(x, str)
    return x


def from_none(x: Any) -> Any:
    assert x is None
    return x


def from_union(fs, x):
    for f in fs:
        try:
            return f(x)
        except:
            pass
    assert False


def from_int(x: Any) -> int:
    assert isinstance(x, int) and not isinstance(x, bool)
    return x


def to_class(c: Type[T], x: Any) -> dict:
    assert isinstance(x, c)
    return cast(Any, x).to_dict()


def from_list(f: Callable[[Any], T], x: Any) -> List[T]:
    assert isinstance(x, list)
    return [f(y) for y in x]


class Property:
    type: str
    unit: Optional[str]
    min: Optional[int]
    max: Optional[int]
    scale: Optional[int]
    step: Optional[int]

    def __init__(self, type: str, unit: Optional[str], min: Optional[int], max: Optional[int], scale: Optional[int], step: Optional[int]) -> None:
        self.type = type
        self.unit = unit
        self.min = min
        self.max = max
        self.scale = scale
        self.step = step

    @staticmethod
    def from_dict(obj: Any) -> 'Property':
        assert isinstance(obj, dict)
        type = from_str(obj.get("type"))
        unit = from_union([from_str, from_none], obj.get("unit"))
        min = from_union([from_int, from_none], obj.get("min"))
        max = from_union([from_int, from_none], obj.get("max"))
        scale = from_union([from_int, from_none], obj.get("scale"))
        step = from_union([from_int, from_none], obj.get("step"))
        return Property(type, unit, min, max, scale, step)

    def to_dict(self) -> dict:
        result: dict = {}
        result["type"] = from_str(self.type)
        result["unit"] = from_union([from_str, from_none], self.unit)
        result["min"] = from_union([from_int, from_none], self.min)
        result["max"] = from_union([from_int, from_none], self.max)
        result["scale"] = from_union([from_int, from_none], self.scale)
        result["step"] = from_union([from_int, from_none], self.step)
        return result


class DeviceSchema:
    mode: str
    code: str
    name: str
    property: Property
    id: int
    type: str
    desc: str
    iconname: Optional[str]

    def __init__(self, mode: str, code: str, name: str, property: Property, id: int, type: str, desc: str, iconname: Optional[str]) -> None:
        self.mode = mode
        self.code = code
        self.name = name
        self.property = property
        self.id = id
        self.type = type
        self.desc = desc
        self.iconname = iconname

    @staticmethod
    def from_dict(obj: Any) -> 'DeviceSchema':
        assert isinstance(obj, dict)
        mode = from_str(obj.get("mode"))
        code = from_str(obj.get("code"))
        name = from_str(obj.get("name"))
        property = Property.from_dict(obj.get("property"))
        id = from_int(obj.get("id"))
        type = from_str(obj.get("type"))
        desc = from_str(obj.get("desc"))
        iconname = from_union([from_str, from_none], obj.get("iconname"))
        return DeviceSchema(mode, code, name, property, id, type, desc, iconname)

    def to_dict(self) -> dict:
        result: dict = {}
        result["mode"] = from_str(self.mode)
        result["code"] = from_str(self.code)
        result["name"] = from_str(self.name)
        result["property"] = to_class(Property, self.property)
        result["id"] = from_int(self.id)
        result["type"] = from_str(self.type)
        result["desc"] = from_str(self.desc)
        result["iconname"] = from_union([from_str, from_none], self.iconname)
        return result

    def get_capability(self) -> Capability:
        ability = Capability(self.code)
        return ability


def deviceschema_from_dict(s: Any) -> DeviceSchema:
    return DeviceSchema.from_dict(s)


def deviceschemas_from_dict(s: Any) -> List[DeviceSchema]:
    return from_list(DeviceSchema.from_dict, s)


def deviceschema_to_dict(x: List[DeviceSchema]) -> Any:
    return from_list(lambda x: to_class(DeviceSchema, x), x)




def order_dict(dictionary):
    return {k: order_dict(v) if isinstance(v, dict) else v
            for k, v in sorted(dictionary.items())}


def scale_number(unscaled, to_min, to_max, from_min, from_max):
    return (to_max - to_min) * (unscaled - from_min) / (from_max - from_min) + to_min


def all_subclasses(cls):
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in all_subclasses(c)])



import asyncio
import codecs
import hashlib
import hmac
import json
import math
import secrets
import time
import uuid

import aiohttp
from aiohttp import ClientSession, TCPConnector

import topgreenerapi.device
from . import util
from .models import location, tuya_device, meta_schema
from .exceptions import TuyaException
from topgreenerapi.models.location import Location
from .logger import logger
from .models.meta_schema import MetaSchema


class TopgreenerCloud:
    API_VERSION = "0.0.1"
    APP_VERSION = "3.8.5"

    def __init__(self, client_id, app_secret, super_secret, app_certhash, device_id=None, region='AZ'):
        self.client_id = client_id
        self.app_secret = app_secret
        self.super_secret = super_secret
        self.app_certhash = app_certhash
        self.sid = None
        self.session = ClientSession(connector=TCPConnector())

        self.keyHmac = self.app_certhash + '_' + self.super_secret + '_' + self.app_secret

        if device_id is None:
            self.device_id = secrets.token_hex(44)
        else:
            self.device_id = device_id

        if region == 'AZ':
            self.region = 'AZ'
            self.endpoint = 'https://a1.tuyaus.com/api.json'
        elif region == 'AY':
            self.region = 'AY'
            self.endpoint = 'https://a1.tuyacn.com/api.json'
        elif region == 'EU':
            self.region = 'EU'
            self.endpoint = 'https://a1.tuyaeu.com/api.json'
        else:
            raise ValueError('Bad region identifier')

    def mobileHash(self, data):
        preHash = hashlib.md5(data.encode('utf-8')).hexdigest()

        return preHash[8:16] + preHash[0:8] + preHash[24:32] + preHash[16:24]

    def signMessage(self, key, strToSign):
        return hmac.new(codecs.encode(key), msg=codecs.encode(strToSign),
                        digestmod=hashlib.sha256).hexdigest()

    async def sendRequest(self, action, data=None, gid=None, requiresSID=True):

        if self.sid is None and requiresSID:
            raise ValueError('You must call login() first!')

        d = str(int(time.time()))
        params = {
            "a": action,
            "deviceId": self.device_id,
            "os": 'Linux',
            "lang": 'en',
            "v": '1.0',
            "clientId": self.client_id,
            "requestId": str(uuid.uuid4()),
            "time": d
        }

        if data is not None:
            logger.debug('Data is %s', data)
            params["postData"] = json.dumps(data)

        if gid is not None:
            params["gid"] = gid

        params["et"] = self.API_VERSION
        params["ttid"] = 'tuya'
        params["appVersion"] = self.APP_VERSION

        if requiresSID:
            params["sid"] = self.sid

        valuesToSign = ['a', 'v', 'lat', 'lon', 'lang', 'deviceId', 'imei',
                        'imsi', 'appVersion', 'ttid', 'isH5', 'h5Token', 'os',
                        'clientId', 'postData', 'time', 'requestId', 'n4h5', 'sid',
                        'sp', 'et']

        sortedPairs = util.order_dict(params)

        strToSign = ''

        for key in sortedPairs:
            if key not in valuesToSign or not params[key]:
                continue
            elif key == 'postData':
                if strToSign:
                    strToSign += '||'
                strToSign += key
                strToSign += '='
                strToSign += self.mobileHash(params[key])
            else:
                if strToSign:
                    strToSign += '||'
                strToSign += key
                strToSign += '='
                strToSign += params[key]

        params["sign"] = self.signMessage(self.keyHmac, strToSign)

        try:
            logger.debug("Sending parameters: %s", params)

            async with self.session.post(self.endpoint, data=params) as resp:
                data = await resp.json()
                logger.debug("Received response: %s", data)

                if "success" not in data or data["success"] is False:
                    raise TuyaException("Request unsuccessful! Params: " + data["errorCode"] + " " + data["errorMsg"])

                return data["result"]

        except aiohttp.ClientError as e:
            logger.error("Error occured! %s", e)
            raise e

    async def get_locations(self) -> list[Location]:
        locations = []
        groups = await self.sendRequest('tuya.m.location.list')
        for group in groups:
            result = location.location_from_dict(group)
            locations.append(result)

        return locations

    async def get_devices(self, locationId) -> list[topgreenerapi.device.Device]:
        devschemas: dict[str, MetaSchema] = {}
        schem = await self.sendRequest('tuya.m.device.ref.info.my.list', gid=locationId)
        schemas = meta_schema.meta_schema_from_dict(schem)
        for dev in schemas:
            devschemas[dev.id] = dev
        deviceArr = await self.sendRequest('tuya.m.my.group.device.list', gid=locationId)

        devices: list[topgreenerapi.device.Device] = []
        for devi in deviceArr:
            tuyaDevice = tuya_device.tuya_device_from_dict(devi)
            device = topgreenerapi.device.Device(tuyaDevice.name, tuyaDevice.dev_id, locationId, tuyaDevice, self)
            device.set_schema(devschemas[tuyaDevice.product_id])
            devices.append(device)

        return devices

    async def register(self, email, password):
        try:
            apiResult = await self.sendRequest('tuya.m.user.email.register',
                                               data={
                                                   "countryCode": self.region,
                                                   "email": email,
                                                   "passwd": hashlib.md5(password).hexdigest()
                                               }, requiresSID=False)
            self.sid = apiResult["sid"]
            return self.sid
        except TuyaException as e:
            if e.errorCode == 'USER_NAME_IS_EXIST':
                return self.login(email, password)

            raise e

    async def login(self, email, password):
        try:
            token = await self.sendRequest('tuya.m.user.email.token.create',
                                           data={
                                               "countryCode": self.region,
                                               "email": email
                                           }, requiresSID=False)

            input = str.encode(hashlib.md5(password.encode('utf-8')).hexdigest())
            e = int(token["exponent"])
            n = int(token["publicKey"])

            keylength = math.ceil(n.bit_length() / 8)
            input_nr = int.from_bytes(input, byteorder='big')
            crypted_nr = pow(input_nr, e, n)
            crypted_data = crypted_nr.to_bytes(keylength, byteorder='big').hex()

            apiResult = await self.sendRequest('tuya.m.user.email.password.login',
                                               data={
                                                   "countryCode": self.region,
                                                   "email": email,
                                                   "passwd": crypted_data,
                                                   "ifencrypt": 1,
                                                   "options": {"group": 1},
                                                   "token": token["token"]
                                               }, requiresSID=False)

            if apiResult["domain"] and apiResult["domain"]["mobileApiUrl"] and not self.endpoint.startswith(
                    apiResult["domain"]["mobileApiUrl"]):
                logger.debug("Changing endpoints after logging: %s -> %s/api.json",
                             self.endpoint, apiResult["domain"]["mobileApiUrl"])

                self.endpoint = apiResult["domain"]["mobileApiUrl"] + '/api.json'
                self.region = apiResult["domain"]["regionCode"]

            self.sid = apiResult["sid"]
            return self.sid
        except Exception as e:
            logger.error("Error while logging in! Trace: %s", e)
            raise e

    async def close(self):
        if self.session:
            await self.session.close()

    async def waitForToken(self, token, devices=1):
        SLEEP_DURATION = 200e-3
        for x in range(200):
            try:
                tokenResult = await self.sendRequest('tuya.m.device.list.token',
                                                     data={
                                                         "token": token
                                                     })

                if tokenResult.length >= devices:
                    return tokenResult

                # wait 200ms and check again
                await asyncio.sleep(SLEEP_DURATION)
            except Exception as e:
                logger.error("Error while waiting for device registration! Stack: %s", e)
                raise e

        raise ValueError("Timed out waiting for device(s) to connect to cloud")



# To use this code, make sure you
#
#     import json
#
# and then, to convert JSON from a string, do
#
#     result = welcome5_from_dict(json.loads(json_string))

from typing import Any, List, TypeVar, Callable, Type, cast


T = TypeVar("T")


def from_str(x: Any) -> str:
    assert isinstance(x, str)
    return x


def from_bool(x: Any) -> bool:
    assert isinstance(x, bool)
    return x


def from_list(f: Callable[[Any], T], x: Any) -> List[T]:
    assert isinstance(x, list)
    return [f(y) for y in x]


def to_class(c: Type[T], x: Any) -> dict:
    assert isinstance(x, c)
    return cast(Any, x).to_dict()


def from_int(x: Any) -> int:
    assert isinstance(x, int) and not isinstance(x, bool)
    return x


class Bic:
    code: str
    selected: bool

    def __init__(self, code: str, selected: bool) -> None:
        self.code = code
        self.selected = selected

    @staticmethod
    def from_dict(obj: Any) -> 'Bic':
        assert isinstance(obj, dict)
        code = from_str(obj.get("code"))
        selected = from_bool(obj.get("selected"))
        return Bic(code, selected)

    def to_dict(self) -> dict:
        result: dict = {}
        result["code"] = from_str(self.code)
        result["selected"] = from_bool(self.selected)
        return result


class PanelConfig:
    bic: List[Bic]

    def __init__(self, bic: List[Bic]) -> None:
        self.bic = bic

    @staticmethod
    def from_dict(obj: Any) -> 'PanelConfig':
        assert isinstance(obj, dict)
        bic = from_list(Bic.from_dict, obj.get("bic"))
        return PanelConfig(bic)

    def to_dict(self) -> dict:
        result: dict = {}
        result["bic"] = from_list(lambda x: to_class(Bic, x), self.bic)
        return result


class SchemaInfo:
    schema_ext: str
    schema: str

    def __init__(self, schema_ext: str, schema: str) -> None:
        self.schema_ext = schema_ext
        self.schema = schema

    @staticmethod
    def from_dict(obj: Any) -> 'SchemaInfo':
        assert isinstance(obj, dict)
        schema_ext = from_str(obj.get("schemaExt"))
        schema = from_str(obj.get("schema"))
        return SchemaInfo(schema_ext, schema)

    def to_dict(self) -> dict:
        result: dict = {}
        result["schemaExt"] = from_str(self.schema_ext)
        result["schema"] = from_str(self.schema)
        return result


class DisplayMsgs:
    dp_switch_led: str
    quickop_dp_switch_led: str
    dp_switch_led_off: str
    quickop_dp_bright_value_unit: str
    quickop_dp_switch_led_off: str
    quickop_dp_bright_value: str
    dp_switch_led_on: str
    quickop_dp_switch_led_on: str

    def __init__(self, dp_switch_led: str, quickop_dp_switch_led: str, dp_switch_led_off: str, quickop_dp_bright_value_unit: str, quickop_dp_switch_led_off: str, quickop_dp_bright_value: str, dp_switch_led_on: str, quickop_dp_switch_led_on: str) -> None:
        self.dp_switch_led = dp_switch_led
        self.quickop_dp_switch_led = quickop_dp_switch_led
        self.dp_switch_led_off = dp_switch_led_off
        self.quickop_dp_bright_value_unit = quickop_dp_bright_value_unit
        self.quickop_dp_switch_led_off = quickop_dp_switch_led_off
        self.quickop_dp_bright_value = quickop_dp_bright_value
        self.dp_switch_led_on = dp_switch_led_on
        self.quickop_dp_switch_led_on = quickop_dp_switch_led_on

    @staticmethod
    def from_dict(obj: Any) -> 'DisplayMsgs':
        assert isinstance(obj, dict)
        dp_switch_led = from_str(obj.get("dp_switch_led"))
        quickop_dp_switch_led = from_str(obj.get("quickop_dp_switch_led"))
        dp_switch_led_off = from_str(obj.get("dp_switch_led_off"))
        quickop_dp_bright_value_unit = from_str(obj.get("quickop_dp_bright_value_unit"))
        quickop_dp_switch_led_off = from_str(obj.get("quickop_dp_switch_led_off"))
        quickop_dp_bright_value = from_str(obj.get("quickop_dp_bright_value"))
        dp_switch_led_on = from_str(obj.get("dp_switch_led_on"))
        quickop_dp_switch_led_on = from_str(obj.get("quickop_dp_switch_led_on"))
        return DisplayMsgs(dp_switch_led, quickop_dp_switch_led, dp_switch_led_off, quickop_dp_bright_value_unit, quickop_dp_switch_led_off, quickop_dp_bright_value, dp_switch_led_on, quickop_dp_switch_led_on)

    def to_dict(self) -> dict:
        result: dict = {}
        result["dp_switch_led"] = from_str(self.dp_switch_led)
        result["quickop_dp_switch_led"] = from_str(self.quickop_dp_switch_led)
        result["dp_switch_led_off"] = from_str(self.dp_switch_led_off)
        result["quickop_dp_bright_value_unit"] = from_str(self.quickop_dp_bright_value_unit)
        result["quickop_dp_switch_led_off"] = from_str(self.quickop_dp_switch_led_off)
        result["quickop_dp_bright_value"] = from_str(self.quickop_dp_bright_value)
        result["dp_switch_led_on"] = from_str(self.dp_switch_led_on)
        result["quickop_dp_switch_led_on"] = from_str(self.quickop_dp_switch_led_on)
        return result


class Shortcut:
    quick_op_dps: List[int]
    fault_dps: List[Any]
    display_dps: List[Any]
    switch_dp: int
    display_msgs: DisplayMsgs

    def __init__(self, quick_op_dps: List[int], fault_dps: List[Any], display_dps: List[Any], switch_dp: int, display_msgs: DisplayMsgs) -> None:
        self.quick_op_dps = quick_op_dps
        self.fault_dps = fault_dps
        self.display_dps = display_dps
        self.switch_dp = switch_dp
        self.display_msgs = display_msgs

    @staticmethod
    def from_dict(obj: Any) -> 'Shortcut':
        assert isinstance(obj, dict)
        quick_op_dps = from_list(from_int, obj.get("quickOpDps"))
        fault_dps = from_list(lambda x: x, obj.get("faultDps"))
        display_dps = from_list(lambda x: x, obj.get("displayDps"))
        switch_dp = from_int(obj.get("switchDp"))
        display_msgs = DisplayMsgs.from_dict(obj.get("displayMsgs"))
        return Shortcut(quick_op_dps, fault_dps, display_dps, switch_dp, display_msgs)

    def to_dict(self) -> dict:
        result: dict = {}
        result["quickOpDps"] = from_list(from_int, self.quick_op_dps)
        result["faultDps"] = from_list(lambda x: x, self.fault_dps)
        result["displayDps"] = from_list(lambda x: x, self.display_dps)
        result["switchDp"] = from_int(self.switch_dp)
        result["displayMsgs"] = to_class(DisplayMsgs, self.display_msgs)
        return result


class UIInfo:
    phase: str
    rn_find: bool
    app_rn_version: str
    ui: str
    type: str

    def __init__(self, phase: str, rn_find: bool, app_rn_version: str, ui: str, type: str) -> None:
        self.phase = phase
        self.rn_find = rn_find
        self.app_rn_version = app_rn_version
        self.ui = ui
        self.type = type

    @staticmethod
    def from_dict(obj: Any) -> 'UIInfo':
        assert isinstance(obj, dict)
        phase = from_str(obj.get("phase"))
        rn_find = from_bool(obj.get("rnFind"))
        app_rn_version = from_str(obj.get("appRnVersion"))
        ui = from_str(obj.get("ui"))
        type = from_str(obj.get("type"))
        return UIInfo(phase, rn_find, app_rn_version, ui, type)

    def to_dict(self) -> dict:
        result: dict = {}
        result["phase"] = from_str(self.phase)
        result["rnFind"] = from_bool(self.rn_find)
        result["appRnVersion"] = from_str(self.app_rn_version)
        result["ui"] = from_str(self.ui)
        result["type"] = from_str(self.type)
        return result


class MetaSchema:
    schema_info: SchemaInfo
    capability: int
    shortcut: Shortcut
    panel_config: PanelConfig
    ui_info: UIInfo
    i18_n_time: int
    attribute: int
    id: str
    category: str
    mesh_category: str
    support_group: bool

    def __init__(self, schema_info: SchemaInfo, capability: int, shortcut: Shortcut, panel_config: PanelConfig, ui_info: UIInfo, i18_n_time: int, attribute: int, id: str, category: str, mesh_category: str, support_group: bool) -> None:
        self.schema_info = schema_info
        self.capability = capability
        self.shortcut = shortcut
        self.panel_config = panel_config
        self.ui_info = ui_info
        self.i18_n_time = i18_n_time
        self.attribute = attribute
        self.id = id
        self.category = category
        self.mesh_category = mesh_category
        self.support_group = support_group

    @staticmethod
    def from_dict(obj: Any) -> 'MetaSchema':
        assert isinstance(obj, dict)
        schema_info = SchemaInfo.from_dict(obj.get("schemaInfo"))
        capability = from_int(obj.get("capability"))
        shortcut = Shortcut.from_dict(obj.get("shortcut"))
        panel_config = PanelConfig.from_dict(obj.get("panelConfig"))
        ui_info = UIInfo.from_dict(obj.get("uiInfo"))
        i18_n_time = from_int(obj.get("i18nTime"))
        attribute = from_int(obj.get("attribute"))
        id = from_str(obj.get("id"))
        category = from_str(obj.get("category"))
        mesh_category = from_str(obj.get("meshCategory"))
        support_group = from_bool(obj.get("supportGroup"))
        return MetaSchema(schema_info, capability, shortcut, panel_config, ui_info, i18_n_time, attribute, id, category, mesh_category, support_group)

    def to_dict(self) -> dict:
        result: dict = {}
        result["schemaInfo"] = to_class(SchemaInfo, self.schema_info)
        result["capability"] = from_int(self.capability)
        result["shortcut"] = to_class(Shortcut, self.shortcut)
        result["panelConfig"] = to_class(PanelConfig, self.panel_config)
        result["uiInfo"] = to_class(UIInfo, self.ui_info)
        result["i18nTime"] = from_int(self.i18_n_time)
        result["attribute"] = from_int(self.attribute)
        result["id"] = from_str(self.id)
        result["category"] = from_str(self.category)
        result["meshCategory"] = from_str(self.mesh_category)
        result["supportGroup"] = from_bool(self.support_group)
        return result


def meta_schema_from_dict(s: Any) -> List[MetaSchema]:
    return from_list(MetaSchema.from_dict, s)


def meta_schema_to_dict(x: List[MetaSchema]) -> Any:
    return from_list(lambda x: to_class(MetaSchema, x), x)



import abc
from strenum import StrEnum
from . import util
from .logger import logger


class TuyaDeviceType(StrEnum):
    LIGHT = "dj"
    CEILING_LIGHT = "xdd"
    AMBIANCE_LIGHT = "fwd"
    STRING_LIGHT = "dc"
    STRIP_LIGHT = "dd"
    MOTION_SENSOR_LIGHT = "gyd"
    CEILING_FAN_LIGHT = "fsd"
    SOLAR_LIGHT = "tyndj"
    DIMMER_LIGHT = "tgq"
    REMOTE_CONTROL_LIGHT = "ykq"
    SPOTLIGHT_LIGHT = "sxd"


class TuyaDeviceCategory(StrEnum):
    LIGHTING = "lighting"
    ELECTRICAL = "electrical"
    LARGE_HOME_APPLIANCE = "large home appliance"
    SMALL_HOME_APPLIANCE = "small home appliance"
    KITCHEN_APPLIANCE = "kitchen appliance"
    SECURITY_AND_VIDEO = "security and video"
    EXERCISE_AND_HEALTH = "exercise and health"
    GATEWAY_CONTROL = "gateway control"
    ENERGY = "energy"
    DIGITAL_ENTERTAINMENT = "digital entertainment"
    OUTDOOR_TRAVEL = "outdoor travel"


class TuyaDeviceTypeInterface(metaclass=abc.ABCMeta):
    @staticmethod
    @abc.abstractmethod
    def get_device_types() -> list[TuyaDeviceType]:
        raise NotImplementedError("User must define get_device_type to use this base class")

    @staticmethod
    @abc.abstractmethod
    def get_device_category() -> TuyaDeviceCategory:
        raise NotImplementedError("User must define get_device_category to use this base class")


class TuyaLightType(TuyaDeviceTypeInterface):
    @staticmethod
    def get_device_types() -> list[TuyaDeviceType]:
        return [TuyaDeviceType.LIGHT]

    @staticmethod
    def get_device_category() -> TuyaDeviceCategory:
        return TuyaDeviceCategory.LIGHTING


class TuyaDimmer(TuyaLightType):

    @staticmethod
    def get_device_types() -> list[TuyaDeviceType]:
        return [TuyaDeviceType.DIMMER_LIGHT]


class TuyaStringLight(TuyaLightType):

    @staticmethod
    def get_device_types() -> list[TuyaDeviceType]:
        return [TuyaDeviceType.STRING_LIGHT]


class TuyaStripLight(TuyaLightType):

    @staticmethod
    def get_device_types() -> list[TuyaDeviceType]:
        return [TuyaDeviceType.STRIP_LIGHT]


def category_code_to_type(category_code: str) -> TuyaDeviceTypeInterface:
    for cls in util.all_subclasses(TuyaDeviceTypeInterface):
        logger.debug("Now checking class: %s, with options: %s", cls, cls.get_device_types())
        if category_code in cls.get_device_types():
            logger.debug("Found! %s, %s", cls, cls())
            return cls()



import os
import unittest
from topgreenerapi import logger
from topgreenerapi import api


class TestApi(unittest.IsolatedAsyncioTestCase):

    def test_message_sign(self):
        key = os.environ["TOPGREENER_key"]
        app_secret = os.environ["TOPGREENER_appsecret"]
        super_secret = os.environ["TOPGREENER_secretkey"]
        app_certhash = os.environ["TOPGREENER_certhash"]
        expectedKey = app_certhash + '_' + super_secret + '_' + app_secret
        api1 = api.TopgreenerCloud(key, app_secret, super_secret, app_certhash)
        toSign = "a=tuya.m.my.group.device.sort.list||appVersion=1.1.1||clientId=uadkfvsv7tpgch9nes7w||deviceId" \
                 "=897757a2214f073ad418da57de36ec408c0b590c8eaa||et=3||lang=en_US||os=Android||postData" \
                 "=d8f5ef5c4e93bfd57fd181e3aa42619f||requestId=46f37579-5bf0-4984-a459-b2d99d50086d||sid" \
                 "=az163879O484027533sYsru32747f67adfaf786dbfb55a2c2541410b||time=1652816648||ttid" \
                 "=sdk_tuya_international@uadkfvsv7tpgch9nes7w||v=1.0"
        expectedSig = "c777ed075a5dd77ccb5645ec445434a3f4d78ec0886034128601de89dda607c7"
        result = api1.signMessage(expectedKey, toSign)
        self.assertEqual(expectedSig, result, 'Signatures did not match expected!')

        toSign = "a=tuya.m.device.product.ref.list||appVersion=1.1.1||clientId=uadkfvsv7tpgch9nes7w||deviceId" \
                 "=897757a2214f073ad418da57de36ec408c0b590c8eaa||et=3||lang=en_US||os=Android||postData" \
                 "=a24b99ff264fca06eca28c2fe0cca478||requestId=4b502814-7d21-42c1-abf8-4a14155872b8||sid" \
                 "=az163879O484027533sYsru32747f67adfaf786dbfb55a2c2541410b||time=1652816648||ttid" \
                 "=sdk_tuya_international@uadkfvsv7tpgch9nes7w||v=1.0"
        expectedSig = "73c690d4074559dc067c45ead09e7e0bb2569a61b88c023ac87fda7358593e68"
        result = api1.signMessage(expectedKey, toSign)
        self.assertEqual(expectedSig, result, 'Signatures did not match expected!')

    async def test_login(self):
        SLEEP_DURATION = 5
        key = os.environ["TOPGREENER_key"]
        app_secret = os.environ["TOPGREENER_appsecret"]
        super_secret = os.environ["TOPGREENER_secretkey"]
        app_certhash = os.environ["TOPGREENER_certhash"]
        api1 = api.TopgreenerCloud(key, app_secret, super_secret, app_certhash)
        sid = await api1.login(os.environ["TOPGREENER_username"], os.environ["TOPGREENER_password"])
        groups = await api1.sendRequest('tuya.m.location.list')
        for group in groups:
            logger.debug("group: %s", group)
            schema = await api1.sendRequest('tuya.m.device.ref.info.my.list', gid=group["groupId"])
            logger.debug("schema!!!!: %s", schema)
            deviceArr = await api1.sendRequest('tuya.m.my.group.device.list', gid=group["groupId"])
            logger.debug("devices: %s", deviceArr)
            for device in deviceArr:
                logger.debug("device data: %s", device)
                logger.debug("group %s device %s devId %s", group["name"], device["name"], device["devId"])

        await api1.close()

    async def test_api_get_dev(self):
        key = os.environ["TOPGREENER_key"]
        app_secret = os.environ["TOPGREENER_appsecret"]
        super_secret = os.environ["TOPGREENER_secretkey"]
        app_certhash = os.environ["TOPGREENER_certhash"]
        api1 = api.TopgreenerCloud(key, app_secret, super_secret, app_certhash)
        sid = await api1.login(os.environ["TOPGREENER_username"], os.environ["TOPGREENER_password"])

        logger.debug("Now fetching locations..")
        locations = await api1.get_locations()
        logger.debug("Locations are %s", locations)

        for location in locations:
            logger.debug("Now fetching devices..")
            devices = await api1.get_devices(location.group_id)
            logger.debug("Devices are: %s", devices)

            for device in devices:
                logger.debug("Device %s capabilities: %s", device.name, device.get_capabilities())

        await api1.close()

    async def test_api_dev_type(self):
        key = os.environ["TOPGREENER_key"]
        app_secret = os.environ["TOPGREENER_appsecret"]
        super_secret = os.environ["TOPGREENER_secretkey"]
        app_certhash = os.environ["TOPGREENER_certhash"]
        api1 = api.TopgreenerCloud(key, app_secret, super_secret, app_certhash)
        sid = await api1.login(os.environ["TOPGREENER_username"], os.environ["TOPGREENER_password"])

        logger.debug("Now fetching locations..")
        locations = await api1.get_locations()
        logger.debug("Locations are %s", locations)

        for location in locations:
            logger.debug("Now fetching devices..")
            devices = await api1.get_devices(location.group_id)
            logger.debug("Devices are: %s", devices)

            for device in devices:
                logger.debug("Device %s capabilities: %s", device.name, device.get_capabilities())
                logger.debug("Device %s is of type: %s", device.name, device.dev_type.get_device_types())

        await api1.close()


if __name__ == '__main__':
    unittest.main()



from typing import List, Any, TypeVar, Callable, Type, cast


T = TypeVar("T")


def from_str(x: Any) -> str:
    assert isinstance(x, str)
    return x


def from_list(f: Callable[[Any], T], x: Any) -> List[T]:
    assert isinstance(x, list)
    return [f(y) for y in x]


def from_int(x: Any) -> int:
    assert isinstance(x, int) and not isinstance(x, bool)
    return x


def from_float(x: Any) -> float:
    assert isinstance(x, (float, int)) and not isinstance(x, bool)
    return float(x)


def from_bool(x: Any) -> bool:
    assert isinstance(x, bool)
    return x


def to_float(x: Any) -> float:
    assert isinstance(x, float)
    return x


def to_class(c: Type[T], x: Any) -> dict:
    assert isinstance(x, c)
    return cast(Any, x).to_dict()


class Location:
    geo_name: str
    rooms: List[Any]
    gmt_modified: int
    group_id: int
    lon: float
    gmt_create: int
    owner_id: int
    uid: str
    background: str
    name: str
    id: int
    lat: float
    status: bool

    def __init__(self, geo_name: str, rooms: List[Any], gmt_modified: int, group_id: int, lon: float, gmt_create: int, owner_id: int, uid: str, background: str, name: str, id: int, lat: float, status: bool) -> None:
        self.geo_name = geo_name
        self.rooms = rooms
        self.gmt_modified = gmt_modified
        self.group_id = group_id
        self.lon = lon
        self.gmt_create = gmt_create
        self.owner_id = owner_id
        self.uid = uid
        self.background = background
        self.name = name
        self.id = id
        self.lat = lat
        self.status = status

    @staticmethod
    def from_dict(obj: Any) -> 'Location':
        assert isinstance(obj, dict)
        geo_name = from_str(obj.get("geoName"))
        rooms = from_list(lambda x: x, obj.get("rooms"))
        gmt_modified = from_int(obj.get("gmtModified"))
        group_id = from_int(obj.get("groupId"))
        lon = from_float(obj.get("lon"))
        gmt_create = from_int(obj.get("gmtCreate"))
        owner_id = int(from_str(obj.get("ownerId")))
        uid = from_str(obj.get("uid"))
        background = from_str(obj.get("background"))
        name = from_str(obj.get("name"))
        id = from_int(obj.get("id"))
        lat = from_float(obj.get("lat"))
        status = from_bool(obj.get("status"))
        return Location(geo_name, rooms, gmt_modified, group_id, lon, gmt_create, owner_id, uid, background, name, id, lat, status)

    def to_dict(self) -> dict:
        result: dict = {}
        result["geoName"] = from_str(self.geo_name)
        result["rooms"] = from_list(lambda x: x, self.rooms)
        result["gmtModified"] = from_int(self.gmt_modified)
        result["groupId"] = from_int(self.group_id)
        result["lon"] = to_float(self.lon)
        result["gmtCreate"] = from_int(self.gmt_create)
        result["ownerId"] = from_str(str(self.owner_id))
        result["uid"] = from_str(self.uid)
        result["background"] = from_str(self.background)
        result["name"] = from_str(self.name)
        result["id"] = from_int(self.id)
        result["lat"] = to_float(self.lat)
        result["status"] = from_bool(self.status)
        return result

    def __repr__(self):
        return str(self.__dict__)


def location_from_dict(s: Any) -> Location:
    return Location.from_dict(s)


def location_to_dict(x: Location) -> Any:
    return to_class(Location, x)



import json
from typing import List
from typing import TYPE_CHECKING

from .models.tuya_device import TuyaDevice

if TYPE_CHECKING:
    from .api import TopgreenerCloud

from . import util
from .capability import Capability
from .logger import logger
from .models import device_schema
from .models.meta_schema import MetaSchema
from .models.device_schema import DeviceSchema
from . import device_type
from .device_type import TuyaDeviceTypeInterface


class Device:
    schema: MetaSchema
    dev_schemas: List[DeviceSchema]
    dev_info: TuyaDevice
    dev_type: TuyaDeviceTypeInterface
    capabilities: dict[Capability, DeviceSchema]

    def __init__(self, name: str, devId: str, gid: str, dev_info: TuyaDevice, api: 'TopgreenerCloud'):
        self.api = api
        self.name = name
        self.devId = devId
        self.gid = gid
        self.dev_info = dev_info
        self.dev_type = device_type.category_code_to_type(self.dev_info.category)
        self.capabilities = {}

    def set_schema(self, schema: MetaSchema):
        self.schema = schema
        self.dev_schemas = device_schema.deviceschemas_from_dict(json.loads(self.schema.schema_info.schema))
        self.capabilities.clear()
        for dev in self.dev_schemas:
            self.capabilities[dev.get_capability()] = dev

    def get_capabilities(self):
        abilities = []
        for capability in self.dev_schemas:
            abilities.append(capability.get_capability())

        return abilities

    def get_device_type(self) -> TuyaDeviceTypeInterface:
        return self.dev_type

    async def set_power_state(self, state: bool):
        if Capability.SWITCH_LED in self.capabilities:
            # then we are good, because the capability exists to do this
            power_ability = self.capabilities[Capability.SWITCH_LED]
            dps = {power_ability.id: state}
            resp = await self.api.sendRequest('tuya.m.device.dp.publish', gid=self.gid,
                                              data={
                                                  "devId": self.devId,
                                                  "gwId": self.devId,
                                                  "dps": dps
                                              })
            logger.debug("Response to power is %s", resp)

    async def set_brightness(self, brightness: int):
        if Capability.BRIGHT_VALUE in self.capabilities:
            bright_ability = self.capabilities[Capability.BRIGHT_VALUE]
            min = bright_ability.property.min
            max = bright_ability.property.max

            nbright = util.scale_number(brightness, min, max, 0, 100)
            dps = {bright_ability.id: nbright}
            resp = await self.api.sendRequest('tuya.m.device.dp.publish', gid=self.gid,
                                              data={
                                                  "devId": self.devId,
                                                  "gwId": self.devId,
                                                  "dps": dps
                                              })
            logger.debug("Response to set brightness %s", resp)





from typing import Any, Optional, Dict, Union, List, TypeVar, Type, cast, Callable
from enum import Enum

T = TypeVar("T")
EnumT = TypeVar("EnumT", bound=Enum)


def from_int(x: Any) -> int:
    assert isinstance(x, int) and not isinstance(x, bool)
    return x


def from_bool(x: Any) -> bool:
    assert isinstance(x, bool)
    return x


def from_str(x: Any) -> str:
    assert isinstance(x, str)
    return x


def from_none(x: Any) -> Any:
    assert x is None
    return x


def from_union(fs, x):
    for f in fs:
        try:
            return f(x)
        except:
            pass
    assert False


def to_enum(c: Type[EnumT], x: Any) -> EnumT:
    assert isinstance(x, c)
    return x.value


def to_class(c: Type[T], x: Any) -> dict:
    assert isinstance(x, c)
    return cast(Any, x).to_dict()


def from_dict(f: Callable[[Any], T], x: Any) -> Dict[str, T]:
    assert isinstance(x, dict)
    return {k: f(v) for (k, v) in x.items()}


def from_list(f: Callable[[Any], T], x: Any) -> List[T]:
    assert isinstance(x, list)
    return [f(y) for y in x]


class DPName:
    pass

    def __init__(self, ) -> None:
        pass

    @staticmethod
    def from_dict(obj: Any) -> 'DPName':
        assert isinstance(obj, dict)
        return DPName()

    def to_dict(self) -> dict:
        result: dict = {}
        return result


class Cdv(Enum):
    EMPTY = ""
    THE_100 = "1.0.0"


class VerSw(Enum):
    THE_314 = "3.1.4"
    THE_433 = "4.3.3"


class Mcu:
    upgrade_status: int
    cdv: Cdv
    ver_sw: VerSw
    is_online: bool
    id: int
    cadv: str
    bv: Optional[str]
    pv: Optional[str]

    def __init__(self, upgrade_status: int, cdv: Cdv, ver_sw: VerSw, is_online: bool, id: int, cadv: str,
                 bv: Optional[str], pv: Optional[str]) -> None:
        self.upgrade_status = upgrade_status
        self.cdv = cdv
        self.ver_sw = ver_sw
        self.is_online = is_online
        self.id = id
        self.cadv = cadv
        self.bv = bv
        self.pv = pv

    @staticmethod
    def from_dict(obj: Any) -> 'Mcu':
        assert isinstance(obj, dict)
        upgrade_status = from_int(obj.get("upgradeStatus"))
        cdv = Cdv(obj.get("cdv"))
        ver_sw = VerSw(obj.get("verSw"))
        is_online = from_bool(obj.get("isOnline"))
        id = from_int(obj.get("id"))
        cadv = from_str(obj.get("cadv"))
        bv = from_union([from_str, from_none], obj.get("bv"))
        pv = from_union([from_str, from_none], obj.get("pv"))
        return Mcu(upgrade_status, cdv, ver_sw, is_online, id, cadv, bv, pv)

    def to_dict(self) -> dict:
        result: dict = {}
        result["upgradeStatus"] = from_int(self.upgrade_status)
        result["cdv"] = to_enum(Cdv, self.cdv)
        result["verSw"] = to_enum(VerSw, self.ver_sw)
        result["isOnline"] = from_bool(self.is_online)
        result["id"] = from_int(self.id)
        result["cadv"] = from_str(self.cadv)
        result["bv"] = from_union([from_str, from_none], self.bv)
        result["pv"] = from_union([from_str, from_none], self.pv)
        return result


class ModuleMap:
    wifi: Mcu
    mcu: Mcu

    def __init__(self, wifi: Mcu, mcu: Mcu) -> None:
        self.wifi = wifi
        self.mcu = mcu

    @staticmethod
    def from_dict(obj: Any) -> 'ModuleMap':
        assert isinstance(obj, dict)
        wifi = Mcu.from_dict(obj.get("wifi"))
        mcu = Mcu.from_dict(obj.get("mcu"))
        return ModuleMap(wifi, mcu)

    def to_dict(self) -> dict:
        result: dict = {}
        result["wifi"] = to_class(Mcu, self.wifi)
        result["mcu"] = to_class(Mcu, self.mcu)
        return result


class TuyaDevice:
    virtual: bool
    dp_name: DPName
    lon: str
    uuid: str
    mac: str
    icon_url: str
    runtime_env: str
    lat: str
    dev_id: str
    dev_key: str
    dp_max_time: int
    product_id: str
    dps: Dict[str, Union[bool, int]]
    ip: str
    active_time: int
    category_code: str
    module_map: ModuleMap
    dev_attribute: int
    name: str
    timezone_id: str
    category: str
    local_key: str

    def __init__(self, virtual: bool, dp_name: DPName, lon: str, uuid: str, mac: str, icon_url: str, runtime_env: str,
                 lat: str, dev_id: str, dev_key: str, dp_max_time: int, product_id: str,
                 dps: Dict[str, Union[bool, int]], ip: str, active_time: int, category_code: str, module_map: ModuleMap,
                 dev_attribute: int, name: str, timezone_id: str, category: str, local_key: str) -> None:
        self.virtual = virtual
        self.dp_name = dp_name
        self.lon = lon
        self.uuid = uuid
        self.mac = mac
        self.icon_url = icon_url
        self.runtime_env = runtime_env
        self.lat = lat
        self.dev_id = dev_id
        self.dev_key = dev_key
        self.dp_max_time = dp_max_time
        self.product_id = product_id
        self.dps = dps
        self.ip = ip
        self.active_time = active_time
        self.category_code = category_code
        self.module_map = module_map
        self.dev_attribute = dev_attribute
        self.name = name
        self.timezone_id = timezone_id
        self.category = category
        self.local_key = local_key

    @staticmethod
    def from_dict(obj: Any) -> 'TuyaDevice':
        assert isinstance(obj, dict)
        virtual = from_bool(obj.get("virtual"))
        dp_name = DPName.from_dict(obj.get("dpName"))
        lon = from_str(obj.get("lon"))
        uuid = from_str(obj.get("uuid"))
        mac = from_str(obj.get("mac"))
        icon_url = from_str(obj.get("iconUrl"))
        runtime_env = from_str(obj.get("runtimeEnv"))
        lat = from_str(obj.get("lat"))
        dev_id = from_str(obj.get("devId"))
        dev_key = from_str(obj.get("devKey"))
        dp_max_time = from_int(obj.get("dpMaxTime"))
        product_id = from_str(obj.get("productId"))
        dps = from_dict(lambda x: from_union([from_int, from_bool], x), obj.get("dps"))
        ip = from_str(obj.get("ip"))
        active_time = from_int(obj.get("activeTime"))
        category_code = from_str(obj.get("categoryCode"))
        module_map = ModuleMap.from_dict(obj.get("moduleMap"))
        dev_attribute = from_int(obj.get("devAttribute"))
        name = from_str(obj.get("name"))
        timezone_id = from_str(obj.get("timezoneId"))
        category = from_str(obj.get("category"))
        local_key = from_str(obj.get("localKey"))
        return TuyaDevice(virtual, dp_name, lon, uuid, mac, icon_url, runtime_env, lat, dev_id, dev_key, dp_max_time,
                          product_id, dps, ip, active_time, category_code, module_map, dev_attribute, name, timezone_id,
                          category, local_key)

    def to_dict(self) -> dict:
        result: dict = {}
        result["virtual"] = from_bool(self.virtual)
        result["dpName"] = to_class(DPName, self.dp_name)
        result["lon"] = from_str(self.lon)
        result["uuid"] = from_str(self.uuid)
        result["mac"] = from_str(self.mac)
        result["iconUrl"] = from_str(self.icon_url)
        result["runtimeEnv"] = from_str(self.runtime_env)
        result["lat"] = from_str(self.lat)
        result["devId"] = from_str(self.dev_id)
        result["devKey"] = from_str(self.dev_key)
        result["dpMaxTime"] = from_int(self.dp_max_time)
        result["productId"] = from_str(self.product_id)
        result["dps"] = from_dict(lambda x: from_union([from_int, from_bool], x), self.dps)
        result["ip"] = from_str(self.ip)
        result["activeTime"] = from_int(self.active_time)
        result["categoryCode"] = from_str(self.category_code)
        result["moduleMap"] = to_class(ModuleMap, self.module_map)
        result["devAttribute"] = from_int(self.dev_attribute)
        result["name"] = from_str(self.name)
        result["timezoneId"] = from_str(self.timezone_id)
        result["category"] = from_str(self.category)
        result["localKey"] = from_str(self.local_key)
        return result


def tuya_device_from_dict(s: Any) -> TuyaDevice:
    return TuyaDevice.from_dict(s)


def tuya_devices_from_dict(s: Any) -> List[TuyaDevice]:
    return from_list(TuyaDevice.from_dict, s)


def tuya_device_to_dict(x: List[TuyaDevice]) -> Any:
    return from_list(lambda x: to_class(TuyaDevice, x), x)
